package proc;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;

import mdp.MDP;

public abstract class Proc {
	// types
    public static final int ZERO = 1;
    public static final int PREFIX = 2;
    public static final int SUM = 3;
    public static final int PARALLEL = 4;
    public static final int PROBSUM = 5;
    public static final int RES = 5;

	public int type = 0;

	protected ArrayList<Transition> trans;
	public abstract ArrayList<Transition> transitions();
	
	public static HashMap<String, Integer> procToId= new HashMap<String, Integer>();
	public static ArrayList<Proc> idToProc = new ArrayList<Proc>();
	public static HashMap<String, Integer> actionToId = new HashMap<String, Integer>();
	public static ArrayList<String> idToAction = new ArrayList<String>();
	private static ArrayList<mdp.Transition[]> mdpTrans = new ArrayList<mdp.Transition[]>();

	public void translateToStates() {
		// pending is a list of processes for which we haven't added states yet
		// procToId maps processes to ids, if a process is there it has been already
		// to pending
		// idToProc maps ids to processes (not needed for this method, we keep them
		// in case we want to "translate" back to processes
		//
		LinkedList<Proc> pending = new LinkedList<Proc>();
		createProcId(this, pending);
		
		while(!pending.isEmpty()) {
			Proc p = pending.pop();
			int pid = createProcId(p, pending);
			
			ArrayList<Transition> pTrans = p.transitions();
			mdp.Transition[] pMdpTrans = new mdp.Transition[pTrans.size()];
			mdpTrans.add(pid, pMdpTrans);
			
			// for each transition of p we add a transition to pMdpTrans
			//
			int i = 0;
			for(Transition t : pTrans) {
				int action = createActionId(t.actionString());
				pMdpTrans[i] = new mdp.Transition(action, t.dist.size());
				
				// for each target state in t, add the corresponding mdp state to pMdpTrans[i]
				//
				int j = 0;
				for(ProbOfProc pp : t.dist) {
					int target = createProcId(pp.process, pending);
					pMdpTrans[i].states[j] = target;
					pMdpTrans[i].probs[j] = pp.prob;
					j++;
				}
				i++;
			}		
		} // while
	}
	
	public static void clearTables() {
		procToId= new HashMap<String, Integer>();
		idToProc = new ArrayList<Proc>();
		actionToId = new HashMap<String, Integer>();
		idToAction = new ArrayList<String>();
		mdpTrans = new ArrayList<mdp.Transition[]>();		
	}
	
	// creates an mdp, using the states/transitions created by previous calls to
	// translateToStates()
	//
	public static MDP createMDP() {
		int stateNo = idToProc.size();
		int actionNo = idToAction.size();
		
		mdp.Transition[][] mdpTransArr = new mdp.Transition[mdpTrans.size()][];
		mdpTrans.toArray(mdpTransArr);
		
		return new MDP(stateNo, actionNo, mdpTransArr);		
	}
			
	private static int createProcId(Proc p, LinkedList<Proc> pending) {
		String s = p.toString();
		Integer id = procToId.get(s);
		if(id == null) {
			id = idToProc.size();
			idToProc.add(id, p);
			procToId.put(s, id);
			
			pending.add(p);
		}
		return id;
	}
	
	private static int createActionId(String action) {
		Integer id = actionToId.get(action);
		if(id == null) {
			id = idToAction.size();
			idToAction.add(id, action);
			actionToId.put(action, id);
		}
		return id;
	}
	
	public int getStateId() {
		return procToId.get(this.toString());
	}
}


